@@ -4,10 +4,12 @@ import datetime
 from typedargs.annotate import docannotate, param, annotated, return_type, context
 from iotile.core.hw.proxy.plugin import TileBusProxyPlugin
 from iotile.core.utilities.packed import unpack
+from iotile.core.utilities.hash_algorithms import KNOWN_HASH_ALGORITHMS
 from iotile.core.exceptions import *
 from iotile.core.hw.reports.report import IOTileReading
 from iotile.core.utilities.console import ProgressBar
 from .lib_controller_types.fw_streamerstatus import StreamerStatus
+from .lib_controller_types.fw_configvariable import ConfigDescriptor
 
 
 @context("SensorGraph")
@@ -131,28 +133,27 @@ class SensorGraphPlugin(TileBusProxyPlugin):
             return parse_binary_descriptor(bindata)
 
     @docannotate
-    def sg_checksum(self):
+    def sg_checksum(self, algorithm="sha256"):
         """ Calculate a checksum over the sensorgraph node data
 
+        Args:
+            algorithm (str): The algorithm to hash the data with
+
         Returns:
             string: The checksum of the sensorgraph
         """
-        try: 
-            from hashlib import sha256
-        except ImportError:
-            raise ExternalError("You must pip install hashlib to use sg_checksum", suggestion="pip install hashlib")
-
-        shasum = sha256()
+        hash_algorithm = KNOWN_HASH_ALGORITHMS[algorithm]
+        nodes_bytes = bytearray()
         nodes = self.count_nodes()
         if nodes > 0:
             for i in range(nodes):
-                shasum.update(self.inspect_node(i,'bin'))
-            return shasum.hexdigest()
+                nodes_bytes += self.inspect_node(i,'bin')
+            return hash_algorithm.calculate(hash_algorithm.algorithm, nodes_bytes)
         else:
             return ""
 
     @docannotate
-    def inspect_streamer(self, index):
+    def inspect_streamer(self, index, fmt=None):
         """Get a descritpion of a streamer using its index.
 
         The description contains the following information:
@@ -165,6 +166,7 @@ class SensorGraphPlugin(TileBusProxyPlugin):
             index (int): The index of the graph streamer that you want to query.
                 If this index does not correspond with an allocated streamer, an
                 ArgumentError will be raised.
+            fmt (string): How to format the return data
 
         Returns:
             DataStreamer show-as string: A description of the streamer at the given index.
@@ -179,8 +181,260 @@ class SensorGraphPlugin(TileBusProxyPlugin):
         if err:
             raise ArgumentError("Invalid streamer index that was not allocated", error_code=err)
 
+        if fmt == 'hex':
+            try:
+                from binascii import hexlify
+            except ImportError:
+                raise ExternalError("You must pip install binascii to use inspect streamer --result_format='hex'", suggestion="pip install binascii")
+
+            return hexlify(desc)
+
+        elif fmt == 'bin':
+            return desc
+
         return parse_binary_descriptor(desc)
 
+    @docannotate
+    def sg_streamer_checksum(self, algorithm="sha256"):
+        """Calculate a checksum over the sensor graph's streamers
+        This function will blindly iterate through streamers until the index
+        of the streamer raises an exception that the streamer doesn't exist.
+        This is how we iterate through all the streamers, since we don't have a
+        RPC to return the number/indices of streamers.
+        Args:
+            algorithm (str): The algorithm to hash the data with
+        Returns:
+            string: The checksum of the sensorgraph
+        """
+
+        streamer_idx = 0
+
+        hash_algorithm = KNOWN_HASH_ALGORITHMS[algorithm]
+        streamers_bytes = bytearray()
+
+        while True:
+            try:
+                streamers_bytes += self.inspect_streamer(streamer_idx, 'bin')
+                streamer_idx += 1
+            except ArgumentError:
+                return hash_algorithm.calculate(hash_algorithm.algorithm,
+                                                streamers_bytes)
+
+    def inspect_config_variable(self, index):
+        """Get a description of a config variable"""
+
+        err, entry = self.rpc(0x2a, 0x0a, index, result_format="L16s")
+
+        if err != 0:
+            raise HardwareError("Could not get config variable information", error_code=err)
+
+        info = ConfigDescriptor(bytearray(entry))
+
+        var_data = bytearray()
+
+        for i in range(0, info.data_length, 16):
+            res = self.rpc(0x2a, 0x0d, index, i, result_type=(0, True))
+            err, = struct.unpack_from("<L", res['buffer'])
+            if err != 0:
+                raise HardwareError("Could not get data for variable", error_code=err, index=index)
+
+            data = res['buffer'][4:]
+            var_data += data
+
+        name, = struct.unpack("<H", var_data[0:2])
+        data = var_data[2:]
+        info.add_name(name)
+        info.add_data(data)
+
+        return info
+
+    @docannotate
+    def sg_config_var_checksum(self, algorithm="sha256", use_ignorelist=True,
+                               ignoreconfigs_list_var=0x2102):
+        """Calculate a checksum over the device's config variables
+        The function iterates through all the config variables and adds it to
+        a list. It will see if there is a config variable that is used to list
+        config variables to exclude during the checksum calculation. It will
+        then build a bytearray of the config variables except those listed
+        in the ignore list, then it will calculate the checksum.
+        Args:
+            algorithm (str): The algorithm to hash the data with
+        Returns:
+            string: The checksum of the config variables
+        """
+
+        hash_algorithm = KNOWN_HASH_ALGORITHMS[algorithm]
+        configs_bytes = bytearray()
+
+        num_vars, = self.rpc(0x2a, 0x0b, result_format="L")
+
+        config_list = []
+        ignoreconfigs_bytes = bytearray()
+
+        for i in range(1, num_vars + 1):
+            config_desc = self.inspect_config_variable(i)
+            config_list.append(config_desc)
+
+            if config_desc.name == ignoreconfigs_list_var:
+                ignoreconfigs_bytes = config_desc.data
+
+        if use_ignorelist:
+            ignore_list = _parse_ignoreconfigs(ignoreconfigs_bytes)
+            for config in config_list:
+                if _is_ignored_config(ignore_list, str(config.target), config.name):
+                    continue
+                configs_bytes += bytes(config.parse_descriptor(), 'utf-8')
+        else:
+            for config in config_list:
+                configs_bytes += bytes(config.parse_descriptor(), 'utf-8')
+
+        return hash_algorithm.calculate(hash_algorithm.algorithm,
+                                        configs_bytes)
+
+    @docannotate
+    def sg_constants_checksum(self, algorithm="sha256"):
+        """Get a description of a constant stream
+        Args:
+            algorithm (str): The algorithm to hash the data with
+        Returns:
+            string: The checksum of the constant streams
+        """
+
+        hash_algorithm = KNOWN_HASH_ALGORITHMS[algorithm]
+        constants_bytes = bytearray()
+        constant_idx = 0
+
+        while True:
+            constant_stream = 'constant ' + str(constant_idx)
+            try:
+                constant_val = self.inspect_virtualstream(constant_stream)
+                constant_stream = "'{}' {}".format(constant_stream, constant_val)
+                constants_bytes += bytes(constant_stream, 'utf-8')
+            except HardwareError:
+                pass
+            except ValidationError as e:
+                break
+
+            constant_idx += 1
+
+        return hash_algorithm.calculate(hash_algorithm.algorithm,
+                                        constants_bytes)
+
+    @docannotate
+    def sg_metadata_checksum(self, algorithm="sha256"):
+        """Calculate a checksum over the device's known metadata information
+        The function calculates the checksum over the only two known and widely
+        used metadata variables, which are the app tag and app version.
+        Args:
+            algorithm (str): The algorithm to hash the data with
+        Returns:
+            string: The checksum of the known metadata
+        """
+
+        hash_algorithm = KNOWN_HASH_ALGORITHMS[algorithm]
+        metadata_bytes = bytearray()
+
+        app_info = self.app_info()
+        app_tag_entry = 'app_tag: ' + str(app_info['tag'])
+        app_version_entry = 'app_version: ' + str(app_info['version_string'])
+
+        metadata_bytes += bytes(app_tag_entry, 'utf-8')
+        metadata_bytes += bytes(app_version_entry, 'utf-8')
+
+        return hash_algorithm.calculate(hash_algorithm.algorithm,
+                                        metadata_bytes)
+
+    @docannotate
+    def calculate_device_checksums(self, algorithm="sha256", use_ignorelist=True):
+        """Returns the checksums of all the device's important actors
+        This function calculates the checksum of checksums of the different
+        configurations that make up the device.
+        Args:
+            algorithm (str): The algorithm to hash the data with
+            use_ignorelist (bool): Whether to ignore variables in checksum
+        Returns:
+            basic_dict: All the checksums related to the device
+        """
+
+        print("Please allow the device about 1 minute to calculate checksums.")
+
+        hash_algorithm = KNOWN_HASH_ALGORITHMS[algorithm]
+        cumulative_bytes = bytearray()
+
+        node_checksum = self.sg_checksum(algorithm=algorithm)
+        streamer_checksum = self.sg_streamer_checksum(algorithm=algorithm)
+        config_checksum = self.sg_config_var_checksum(algorithm=algorithm, use_ignorelist=use_ignorelist)
+        constants_checksum = self.sg_constants_checksum(algorithm=algorithm)
+        metadata_checksum = self.sg_metadata_checksum(algorithm=algorithm)
+
+        # The following order is important and must match how iotile-sgcompile
+        # calculates its checksum.
+        # See iotilesensorgraph/iotile/sg/graph.py::add_checksum
+        cumulative_bytes += bytes(node_checksum, 'utf-8')
+        cumulative_bytes += bytes(streamer_checksum, 'utf-8')
+        cumulative_bytes += bytes(config_checksum, 'utf-8')
+        cumulative_bytes += bytes(constants_checksum, 'utf-8')
+        cumulative_bytes += bytes(metadata_checksum, 'utf-8')
+
+        device_checksum = hash_algorithm.calculate(hash_algorithm.algorithm,
+                                                   cumulative_bytes)
+        return {
+            'nodes': node_checksum,
+            'streamers': streamer_checksum,
+            'configs': config_checksum,
+            'constants': constants_checksum,
+            'metadata': metadata_checksum,
+            'device': device_checksum
+        }
+
+    @docannotate
+    def compare_device_checksum(self, checksum_address=0x2100, hash_algorithm_address=0x2101):
+        """Compares the calculated checksum with the programmed checksum
+        Returns:
+            bool: Whether the calculated checksum matches the programmed one
+        """
+
+        programmed_checksum = self.get_programmed_checksum(checksum_address=checksum_address)
+        hash_algorithm = self.get_hash_algorithm(hash_algorithm_address=hash_algorithm_address)
+
+        calculated_checksum = self.calculate_device_checksums(algorithm=hash_algorithm)['device']
+
+        return True if calculated_checksum == programmed_checksum else False
+
+    @docannotate
+    def get_programmed_checksum(self, checksum_address=0x2100):
+        """Returns the checksum that was programmed into its config var
+        Returns:
+            string: The checksum programmed into the variable
+        """
+
+        num_vars, = self.rpc(0x2a, 0x0b, result_format="L")
+
+        for i in range(1, num_vars + 1):
+            config_desc = self.inspect_config_variable(i)
+
+            if config_desc.name == checksum_address:
+                return config_desc.data.decode('utf-8')
+
+        return
+
+    @docannotate
+    def get_hash_algorithm(self, hash_algorithm_address=0x2101):
+        """Returns the hash algorithm that was programmed into its config var
+        Returns:
+            string: The hash algorithm used
+        """
+
+        num_vars, = self.rpc(0x2a, 0x0b, result_format="L")
+
+        for i in range(1, num_vars + 1):
+            config_desc = self.inspect_config_variable(i)
+
+            if config_desc.name == hash_algorithm_address:
+                return config_desc.data.decode('utf-8')
+
+        return
+
     @param("node", "fw_graphnode", desc="SensorGraphNode object to add to the sensor graph")
     def add_node(self, node):
         """
@@ -822,3 +1076,31 @@ class SensorGraphPlugin(TileBusProxyPlugin):
         result['value0'] = value0
         result['value1'] = value1
         return result
+
+
+def _parse_ignoreconfigs(ignoreconfigs_bytes):
+    """"""
+
+    ignoreconfigs_str = ignoreconfigs_bytes.decode('utf-8')
+    configs_ignorelist = []
+
+    if ignoreconfigs_str[0] != '{' and ignoreconfigs_str[-1] != '}':
+        ignoreconfigs_str = '{' + ignoreconfigs_str + '}'
+
+    ignoreconfigs_dict = eval(ignoreconfigs_str)
+
+    for target, config_ids in ignoreconfigs_dict.items():
+        for config_id in config_ids:
+            configs_ignorelist.append((target, config_id))
+
+    return configs_ignorelist
+
+
+def _is_ignored_config(ignore_configs, slot, config_var):
+    """Checks if an entry is in the list of variables to ignore"""
+
+    for ignore_config in ignore_configs:
+        if slot == ignore_config[0] and config_var == ignore_config[1]:
+            return True
+
+    return False
\ No newline at end of file
